#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "bitmap.h"
#include "argv.h"

#define ImageWidth 12800
#define ImageHeight 1024
#define SquareSize 6
#define NumImages 100000
#define SPACE 3

int clearImage(tColorImage *image) {
  unsigned int x,y;
  tColorPixel tmp;
      
  for(x=0;x< ImageWidth;x++){
    for(y=0; y < ImageHeight;y++){
	    tmp[0] = 255;
	    tmp[1] = 255;
	    tmp[2] = 255;
	    BMP_Write_Color_Pixel(image, y, x, tmp);
   }
  }
  return 0;

}

int drawSquare (tColorImage *image, int x, int y) {
  tColorPixel tmp, p;
  int posx, posy;
  int mySize;

  tmp[0] = 0;
  tmp[1] = 0;
  tmp[2] = 0;

  mySize = SquareSize + rand() % 5;

  for (posx= x - mySize - SPACE; posx <= x + mySize + SPACE ;posx++) {
	  for (posy= y - mySize - SPACE; posy <= y+ mySize + SPACE; posy++) {
	     if ((posx <0) || (posy < 0) || (posx >= ImageWidth) || (posy >= ImageHeight)) 
		     return -1;
	     BMP_Read_Color_Pixel(image, posy, posx, &p);
	     if ((p[0] != 255) && (p[1]!= 255) && (p[2]!=255))
		     return -1;
	  }
  }
  for (posx= x - mySize; posx <= x + mySize;posx++) {
	  for (posy= y - mySize;posy <= y + mySize;posy++) {
	      BMP_Write_Color_Pixel(image, posy, posx, tmp);
		    
          }
  }
  
  return 0;
}



int draw_figure(tColorImage *c1,unsigned int x,unsigned int y, FILE* f){
    int res; 
	       
    res = drawSquare (c1,x,y);
    if (!res)
	fprintf(f,"%d %d\n",x,y);
    return res;
}


void uniform_distrib(tColorImage *c1, char *fname){
    FILE *f = fopen(fname,"w+");
    int drawed = 0,  error = 0;
    unsigned int x,y;
    while( drawed < NumImages && error < NumImages * 20) {	
	x = rand() % ImageWidth;
	y = rand() % ImageHeight;
	if (draw_figure(c1,x,y,f) == 0)  
	    drawed++;
	else 
	    error++;
    }
    printf ("Drawed %d objects\n", drawed);
    fclose(f);
}


int draw_between(tColorImage *c1, unsigned int y1, unsigned int y2, int max, FILE* f) {
    int drawed = 0;
    int error = 0;
    unsigned int x,y;
    while (drawed < max &&  error < NumImages * 20) {
	x = rand() % ImageWidth;
	y = rand() % (y2 - y1);
	if (draw_figure(c1,x,y + y1,f) == 0)  
	    drawed++;
	else 
	    error++;

    }
    printf ("Drawed %d objects between y1=%d and y2=%d\n", drawed, y1, y2);
    return drawed;
}

void geometric_distrib(tColorImage *c1, int nr_proc, double rate, char *fname) {
    FILE *f = fopen(fname,"w+");
    unsigned int rows_per_proc = c1->rows / nr_proc;
    int i, max = NumImages, total_drawed = 0;
    for (i=0;i<nr_proc;i++) {
	max = draw_between(c1, i*rows_per_proc,(i+1)* rows_per_proc, max, f); 
	total_drawed += max;
	max = (int)((double)max * rate);
    }
    printf ("**************\nDrawed objects (total)=%d\n", total_drawed);
    fclose(f);

}

void read_centers_from_file(char *fname) {
    FILE *f = fopen(fname,"r");
    int x,y;
    //   int i=0;

    while(!feof(f)){
	fscanf(f,"%5d",&y);
	fscanf(f,"%5d",&x);
/*	if (i++ < 10) // Print the first 10 to verify
	printf("x=%d, y=%d\n",x,y);*/
    }

    fclose(f);

}
char fname[256];
int distrib;
int nr_procs = 8;
double ratio = .9;

static argv_t args[] = {
  { ARGV_MAND, NULL, ARGV_CHAR_P, fname, "img-file","Image file to be generated" },
  { ARGV_MAND, NULL, ARGV_INT, &distrib, "object-distribution","0:uniform 1:geometric" },
  { 'n', NULL, ARGV_INT, &nr_procs, "geom-prog-size","default:8" },
  { 'r', NULL, ARGV_DOUBLE, &ratio, "geom-ratio","default:.9" }, 
  { ARGV_LAST }
};


int main( int argc, char *argv[] )
{
  tColorImage c1;
  argv_help_string = "Generating bmp figure.";
  argv_usage_type = ARGV_USAGE_LONG;
  argv_process(args, argc, argv);

  sprintf(fname,"%s.centers",argv[1]);
 
  BMP_Create_Color_Image( &c1, ImageHeight, ImageWidth );
  clearImage (&c1);
  if (distrib == 0)
      uniform_distrib(&c1,fname);
  else 
      geometric_distrib(&c1,nr_procs,ratio,fname);
  BMP_Write_Color_Image( argv[1], &c1);
  BMP_Free_Color_Image(&c1);


  read_centers_from_file(fname);
  return 0;

}
